
(* ocamlopt -o coquelicote lex.ml coquelicote_txt.ml *)

type intui_type =
  | Omega
  | Arrow of string * intui_type * intui_type
  | Witness of intui_expr
and  intui_expr =
  | Forall of string * intui_type * intui_expr
  | Prop of string 
  | Var of string 
  | Lambda of string * intui_type * intui_expr 
  | Apply of intui_expr * intui_expr

type intui_context = (string * intui_type) list

let rec from_arrow t =
  match t with
  | Arrow(x,a,b) -> 
     ( match from_arrow b with
     | Witness e -> Witness(Forall(x,a,e))
     | _ -> t ) 
  | _ -> t

let rec to_type e (c: intui_context) =
  match e with
  | Forall(x,t,e) ->
      if to_type e ((x,t)::c) = Omega then Omega
      else failwith "type error: expected a propositional value"
  | Prop e  ->
      Omega
  | Var x  ->
      ( try List.assoc x c with
      | Not_found -> failwith ("unkown variable: " ^ x)
      )
  | Lambda(x,t,e) ->
      from_arrow (Arrow(x,t,from_arrow (to_type e ((x,t)::c))))
  | Apply(f,arg) ->
      ( match to_type f c with
      | Arrow(x,a,b) when a = to_type arg c -> from_arrow b  
      | Arrow _ -> failwith "type error: bad arrow source"
      | Witness(Forall(x,t,e)) when t = to_type arg c -> Witness e 
      | Witness _ -> failwith "type error: bad "
      | _ -> failwith "type error: expected a functional value" )

let rec print_type t =
  match t with
  | Omega ->
      print_char 'O'; 
  | Arrow(x,a,b) ->
      if x = "_" then
        print_type a
      else begin
        print_char '(';
        print_string x;
        print_char ':'; 
        print_type a; 
        print_char ')';
      end; 
      print_string " -> "; 
      print_type b; 
  | Witness(e) ->
      print_char 'W'; 
      print_char '('; 
      print_expr e; 
      print_char ')';
and print_expr e =
  match e with
  | Forall(x,t,e) ->
      print_char '?'; 
      print_string x; 
      print_char ':'; 
      print_type t; 
      print_char ' '; 
      print_expr e; 
  | Prop e  ->
      print_string e
  | Var x  ->
      print_string x
  | Lambda(x,t,e) ->
      print_char '('; 
      print_string x; 
      print_char ':'; 
      print_type t; 
      print_char ')'; 
      print_string " |-> "; 
      print_expr e;
  | Apply(f,a) ->
      print_expr f;
      print_char '('; 
      print_expr a;
      print_char ')'
      ;;

print_endline "Coquelicote Text v0.1";

open Lex;;

while true do

  print_string  "# "; flush stdout;

  let lex = Lex.make stdin in

  let rec demand_expr () =
  match lex with
  | l when lex.granted_mono '?' ->
      let x = lex.demand_name () in
      lex.demand_mono ':';
      let t = demand_type () in
      Forall(x,t,demand_expr ())
  | l when lex.granted_mono '(' ->
      let x = lex.demand_name () in
      lex.demand_mono ':';
      let t = demand_type () in
      lex.demand_mono ')';
      lex.demand_poly "|->";
      Lambda(x,t,demand_expr ())
  | l ->
      let x = lex.demand_name () in
      if 'A' <= x.[0] && x.[0] <= 'Z' then
        Prop(x)
      else
        demand_apply
        ( match lex with
        | l when lex.granted_mono ':' ->
            let t = demand_type () in
            lex.demand_poly "|->";
            Lambda(x,t,demand_expr ())
        | l -> Var(x) )
  and demand_apply f =
    if not (lex.granted_mono '(') then f else
    let arg = demand_expr () in
    lex.demand_mono ')';
    demand_apply (Apply(f,arg)) 
  and demand_type () =
    let t =
    match lex with
    | l when lex.granted_word "O" ->
        Omega
    | l when lex.granted_word "W" ->
        lex.demand_mono '(';
        let e = demand_expr () in      
        lex.demand_mono ')';
        Witness(e)
    | l ->
        lex.demand_mono '(';
        let t = demand_type () in      
        lex.demand_mono ')'; t
    in
    if not (lex.granted_poly "->") then t
    else (Arrow("_",t,demand_type ()))

  in

  try
    if lex.promised_word "quit" then exit 0;
    print_string  ": ";
    print_type (to_type (demand_expr ()) []);
    if not (lex.promised_mono '.') then lex.demand_mono '.';
    print_newline (); 
  with
  | Demand_denied (p,d) -> print_position p; print_demand d
  | Failure msg -> print_endline msg
  
done
